# netbots.tcl v3.35 (16 June 2000)
# copyright (c) 1998-2000 by slennox <slennox@egghelp.org>
# slennox's eggdrop page - http://www.egghelp.org/

## netbots.tcl core script ##

## Please read netbots.txt and components.txt before using this script ##
## Settings must be configured in netset.tcl ##

set nb_ver "v3.35" ; set nb_numver "3350"
putlog "Loading netbots.tcl $nb_ver by slennox.."

if {[string trimleft [lindex [split $version] 1] 0] < 1031700} {
  die "*** netbots.tcl $nb_ver requires eggdrop 1.3.17 or later. Please upgrade your eggdrop, or remove the script."
}

set nb_script [info script]
set nb_dir [file dirname $nb_script]
set nb_setscript $nb_dir/netset.tcl
if {![file exists $nb_setscript]} {
  putlog "*** Couldn't find netbots.tcl settings file '$nb_setscript'."
  return
}

if {$numversion < 1040200} {
  if {[info commands isop] != ""} {
    proc wasop {nick chan} {
      return [isop $nick $chan]
    }
  }
  if {$numversion < 1032500} {
    proc islinked {bot} {
      if {[lsearch -exact [string tolower [bots]] [string tolower $bot]] == -1} {return 0}
      return 1
    }
    if {$numversion < 1032400} {
      if {[info commands onchan] != ""} {
        proc botonchan {chan} {
          global botnick
          if {![validchan $chan]} {
            error "illegal channel: $chan"
          } elseif {![onchan $botnick $chan]} {
            return 0
          }
          return 1
        }
      }
      if {[info commands putserv] != ""} {
        proc putquick {text} {
          putserv $text
        }
      }
    }
  }
}

proc isbotnetnick {nick} {
  global botnet-nick
  if {[string tolower $nick] == [string tolower ${botnet-nick}]} {return 1}
  return 0
}

proc slindex {string n} {
  return [lindex [split $string] $n]
}

proc stl {string} {
  return [string tolower $string]
}

if {((![info exists botnet-nick]) || (${botnet-nick} == ""))} {
  set botnet-nick $nick
}

proc nb_set {bots setting value} {
  global botnet-nick [lindex [split $setting "("] 0]
  if {[lsearch -exact [split [string tolower $bots] ,] [string tolower ${botnet-nick}]] != -1} {
    set $setting $value
  }
}

proc nb_berror {file} {
  global nb_error
  if {[info commands nb_sendcmd] != ""} {
    nb_sendcmd * rbroadcast "*** WARNING: error found when loading $file:"
    nb_sendcmd * rbroadcast $nb_error
  }
}

if {[info exists nb_ctrlbots]} {
  unset nb_ctrlbots
}

if {[catch {source $nb_setscript} nb_error]} {
  putlog "*** WARNING: error found when loading $nb_setscript:"
  putlog $nb_error
  utimer 3 [split "nb_berror $nb_setscript"]
  return
}

if {[info exists nb_helpidx]} {
  unset nb_helpidx
}
if {[info exists nb_help]} {
  unset nb_help
}
if {[info exists nb_netcmds]} {
  unset nb_netcmds
}

proc nb_dccnethelp {hand idx arg} {
  global nb_component nb_help nb_helpidx nb_owner owner
  putcmdlog "#$hand# nethelp $arg"
  set command [string tolower [string trimleft [lindex [split $arg] 0] "."]]
  set chan [lindex [console $idx] 0]
  if {$command == ""} {
    if {(([matchattr $hand n]) && ((!$nb_owner) || ([lsearch -exact [split [string tolower $owner] ", "] [string tolower $hand]] != -1)))} {
      putidx $idx "netbots.tcl commands"
      if {$nb_owner} {
        putidx $idx " For permanent owners:"
      } else {
        putidx $idx " For owners:"
      }
      putidx $idx "  netbots       netcontrol    netinfo       netshell"
      putidx $idx "  netserv       netpass       netsave       nethash"
      putidx $idx "  netsay        netact        netnotice     netdump"
      putidx $idx "  netnick       netjump       netjoin       netpart"
      putidx $idx "  netchanset    netupdate     components"
      putidx $idx " "
      set found 1
    }
    if {[info exists nb_helpidx]} {
      foreach procname $nb_helpidx {
        if {[$procname $hand $chan $idx]} {
          set found 1
        }
      }
    }
    if {![info exists found]} {
      putidx $idx "What?  You need '.help'"
    }
  } else {
    if {[info exists nb_help($command)]} {
      if {[$nb_help($command) $hand $chan $idx $command]} {
        return 0
      }
    }
    putidx $idx "What?  You need '.nethelp'"
  }
  return 0
}

proc nb_help {hand chan idx cmd} {
  global nb_owner owner
  if {((![matchattr $hand n]) || (($nb_owner) && ([lsearch -exact [split [string tolower $owner] ", "] [string tolower $hand]] == -1)))} {return 0}
  switch -exact -- $cmd {
    "nethelp" {
      putidx $idx "# nethelp \[command\]"
      putwrap $idx 3 "Displays information about the specified command. If no command is specified, this will display a list of all the available netbot commands."
      return 1
    }
    "netbots" {
      putidx $idx "# netbots \[add|remove <handle1,handle2,etc>\]"
      putwrap $idx 3 "This command lets you add bots to or remove bots from the netbot list. You can specify a list of bots to add/remove by separating their handles with commas or spaces. It automatically gives the specified bot(s) the netbot flag. The bot(s) will only send commands to and receive commands from other netbots. If no options are specified, this command lists all current netbots."
      putwrap $idx 3 "Note: this command is not used if nb_flag is set to 'b' or 'all' in the script's options."
      return 1
    }
    "control" -
    "netcontrol" {
      putidx $idx "# netcontrol/control \[*|group|bot1,bot2,etc\]"
      putwrap $idx 3 "Allows you to select which netbots to control with netbot commands. Specifying '*' will let you control the current bot and all other netbots. Specifying a group of netbots (as defined in the config file using 'nb_group') will make netbot commands control that group only (only including the current bot if it's part of the group). Specifying a list of bots (separated by commas or spaces) will make netbot commands control that list of bots only (only including the current bot if it's listed)."
      putwrap $idx 3 "The netcontrol command is effective until you sign off from the bot. So, for example, if you use .netcontrol \[group\], any netbot commands you use will control that group, until you either use .netcontrol to change the bots to control, or sign off from the bot."
      putwrap $idx 3 "If no arguments are specified, the command will list the netbots you're currently controlling, as well as any netbot groups available to control."
      return 1
    }
    "netinfo" {
      putidx $idx "# netinfo"
      putwrap $idx 3 "Displays information about the version of netbots.tcl each netbot is running, and the components each has loaded."
      return 1
    }
    "netshell" {
      putidx $idx "# netshell"
      putwrap $idx 3 "Displays brief information about the shell each netbot is running on, including operating system, uptime, and load averages."
      return 1
    }
    "netserv" {
      putidx $idx "# netserv"
      putwrap $idx 3 "Displays the server each netbot is using and time connected."
      return 1
    }
    "netpass" {
      putidx $idx "# netpass \[password \[handle\]\]"
      putwrap $idx 3 "This command works in different ways depending on which options are specified. If both a password and handle are specified, this will change the password for that particular user across the botnet. If only a password is specified, this changes the 'netbot password' (the userfile password each netbot has for all other netbots) to the specified password. If no options are specified, this changes the netbot password to a random alpha-numeric string."
      putwrap $idx 3 "Note: if you set nb_flag to 'b' or 'all' in the script's options this command is only partially enabled - you can change the password for a particular user, but cannot change the netbot password."
      putidx $idx "# netpass -check"
      putwrap $idx 3 "Displays a list of users without a password set for each netbot."
      return 1
    }
    "netsave" {
      putidx $idx "# netsave"
      putwrap $idx 3 "Makes all netbots perform a userfile/chanfile save."
      return 1
    }
    "nethash" {
      putidx $idx "# nethash"
      putwrap $idx 3 "Makes all netbots perform a rehash."
      return 1
    }
    "netsay" {
      putidx $idx "# netsay <channel|nick> <message>"
      putwrap $idx 3 "Makes all netbots say the specified message to the specified channel or nick."
      return 1
    }
    "netact" {
      putidx $idx "# netact <channel|nick> <message>"
      putwrap $idx 3 "Makes all netbots perform the specified action to the specified channel or nick."
      return 1
    }
    "netnotice" {
      putidx $idx "# netnotice <channel|nick> <notice>"
      putwrap $idx 3 "Makes all netbots send the specified notice to the specified channel or nick."
      return 1
    }
    "netdump" {
      putidx $idx "# netdump <text>"
      putwrap $idx 3 "Makes all netbots send the specified text to the server. Basically the same as eggdrop's .dump command. Be careful when using this."
      return 1
    }
    "netnick" {
      putidx $idx "# netnick \[-rand\]"
      putwrap $idx 3 "Makes all netbots change their nicknames to their alternate nickname. If -rand is specified, it instead changes their nicks to the first three letters of their nick with a random number between 10 and 999999 added on the end."
      putidx $idx "# netnick <bot1:newnick bot2:newnick botN:newnick>"
      putwrap $idx 3 "Makes each specified netbot change its nickname to the specified new nickname. For example, if you have a bot named 'Tom' and another named 'Harry', and would like to change their nicks to 'Dumb' and 'Dumber' respectively, you would use: .netnick Tom:Dumb Harry:Dumber"
      putidx $idx " "
      putwrap $idx 3 "Tip: to make the bots switch back to their ordinary nicks, use nethash."
      return 1
    }
    "netjump" {
      putidx $idx "# netjump <bot|-all> \[server\]"
      putwrap $idx 3 "Makes netbots jump to the next server in their server list. If -all is specified, all currently controlled netbots will jump. If a bot is specified, only that bot will jump. If a server is specified, the bot(s) will jump to the specified server. You can specify the server port/password in server:port:password format."
      return 1
    }
    "netjoin" {
      putidx $idx "# netjoin <channel> \[key\]"
      putwrap $idx 3 "Adds the specified channel to all netbots. If the channel has a key, you should specify the key so that the bots can join."
      return 1
    }
    "netpart" {
      putidx $idx "# netpart <channel>"
      putwrap $idx 3 "Removes the specified channel from all netbots."
      putwrap $idx 3 "Tip: if your netbots are 1.3.28+, you can take advantage of the 'inactive' channel setting to make all netbots temporarily part a channel (without losing channel-specific flags, etc.). Simply use '.netchanset <channel> +inactive'."
      return 1
    }
    "netchanset" {
      putidx $idx "# netchanset <channel|-all> \[settings\]"
      putwrap $idx 3 "Allows you to change a channel's settings on all netbots. This command works in basically the same way as eggdrop's .chanset command. Specifying -all instead of a channel will make the chanset take effect on all channels. If no settings are specified, the channel's settings are changed to the default set in the script's options."
      putidx $idx "# netchanset -reload"
      putwrap $idx 3 "Resets all channel settings as set in the config file. Normally, the chanfile will override the config file's channel settings, preventing any changes you make from taking effect unless you use the .chanset command. This command lets you make changes in the bot config files and reload the channel settings from there, without the need to disable the chanfile."
      return 1
    }
    "netupdate" {
      putidx $idx "# netupdate \[-config|-motd|-tbanner|-settings|-scripts|-everything\] \[bot\]"
      putwrap $idx 3 "You should typically use this command on the hub bot, after updating config files, motd, telnet-banner, netbots.tcl or a component script, or settings file on the hub's shell. The first option specifies which file(s) is updated."
      putidx $idx "    -config sends bot config files."
      putidx $idx "    -motd sends motd files."
      putidx $idx "    -tbanner sends telnet-banner files."
      putidx $idx "    -settings sends the netbots settings file."
      putidx $idx "    -scripts sends netbots.tcl and component scripts."
      putidx $idx "    -everything sends all of the above."
      putwrap $idx 3 "If no switch is specified, a settings/scripts update is performed. If a bot is specified, the requested file is sent to that bot only, instead of updating all netbots."
      putidx $idx "# netupdate -file <filename> \[bot\]"
      putwrap $idx 3 "Allows you to update a specific file across the botnet. This should only be used with text files (e.g. Tcl scripts). Examples:"
      putidx $idx "    netupdate -file awesome.tcl"
      putidx $idx "     Sends 'awesome.tcl' to all netbots. If the file doesn't exist, it"
      putidx $idx "     will be created. If it exists, it will be overwritten."
      putidx $idx "    netupdate -file scripts/groovy.tcl  OR"
      putidx $idx "    netupdate -file ./scripts/groovy.tcl"
      putidx $idx "     Sends 'groovy.tcl' to all netbots. The netbots will place it in"
      putidx $idx "     the 'scripts' directory. If the directory doesn't exist, it will"
      putidx $idx "     be created."
      putidx $idx " "
      putwrap $idx 3 "See the script documentation for more information about this feature."
      return 1
    }
    "components" {
      putidx $idx "# components"
      putwrap $idx 3 "Displays a list of loaded components and a list of available components."
      return 1
    }
  }
  return 0
}

set nb_help(netbots) "nb_help"
set nb_help(nethelp) "nb_help"
set nb_help(control) "nb_help"
set nb_help(netcontrol) "nb_help"
set nb_help(netinfo) "nb_help"
set nb_help(netshell) "nb_help"
set nb_help(netserv) "nb_help"
set nb_help(netpass) "nb_help"
set nb_help(netsave) "nb_help"
set nb_help(nethash) "nb_help"
set nb_help(netsay) "nb_help"
set nb_help(netact) "nb_help"
set nb_help(netnotice) "nb_help"
set nb_help(netdump) "nb_help"
set nb_help(netnick) "nb_help"
set nb_help(netjump) "nb_help"
set nb_help(netjoin) "nb_help"
set nb_help(netpart) "nb_help"
set nb_help(netchanset) "nb_help"
set nb_help(netupdate) "nb_help"

proc nb_dcccomponents {hand idx arg} {
  global nb_components nb_dir
  putcmdlog "#$hand# components $arg"
  putidx $idx "Loaded components: [join $nb_components ", "]"
  if {[catch {glob -- $nb_dir/*.tcl} scripts]} {
    putlog "Unable to list component scripts in '$nb_dir'."
    return 0
  }
  foreach script $scripts {
    set script [file tail $script]
    if {$script != "netset.tcl"} {
      lappend available [file rootname $script]
    }
  }
  putidx $idx "Available components: [join [lsort $available] ", "]"
  return 0
}

proc nb_dccnetbots {hand idx arg} {
  global nb_control nb_flag
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netbots $arg"
  if {(($nb_flag == "b") || ($nb_flag == "all"))} {
    putidx $idx "This command is not used if nb_flag is set to 'b' or 'all'." ; return 0
  }
  set command [lindex [split $arg] 0] ; set bothands [lrange [split $arg] 1 end]
  if {$command == "add"} {
    if {$bothands == ""} {
      putidx $idx "You must specify a handle." ; return 0
    }
    set added ""
    foreach bothand $bothands {
      if {![validuser $bothand]} {
        putidx $idx "$bothand is not a valid user." ; continue
      }
      if {![matchattr $bothand b]} {
        putidx $idx "$bothand is not a valid bot." ; continue
      }
      if {[nb_control]} {
        chattr $bothand +$nb_flag
      }
      nb_sendcmd $nb_control netbots "$hand add $bothand"
      lappend added $bothand
    }
    putidx $idx "Added [join $added ", "] to netbot list."
  } elseif {$command == "remove"} {
    if {$bothands == ""} {
      putidx $idx "You must specify a handle." ; return 0
    }
    set removed ""
    foreach bothand $bothands {
      if {![validuser $bothand]} {
        putidx $idx "$bothand is not a valid user." ; continue
      }
      if {![matchattr $bothand b]} {
        putidx $idx "$bothand is not a valid bot." ; continue
      }
      nb_sendcmd $nb_control netbots "$hand remove $bothand"
      if {[nb_control]} {
        chattr $bothand -$nb_flag
      }
      lappend removed $bothand
    }
    putidx $idx "Removed [join $removed ", "] from netbot list."
  } elseif {$command == ""} {
    set botlist "" ; set offline ""
    foreach netbot [userlist $nb_flag] {
      lappend botlist $netbot
      if {((![islinked $netbot]) && (![isbotnetnick $netbot]))} {
        lappend offline $netbot
      }
    }
    if {$botlist == ""} {
      putidx $idx "There are no netbots." ; return 0
    }
    putidx $idx "Current netbots: [join $botlist ", "]."
    if {$offline != ""} {
      putidx $idx "Offline or not linked: [join $offline ", "]."
    } else {
      putidx $idx "All netbots are online and linked."
    }
  } else {
    putidx $idx "Usage: netbots \[add|remove <handle>\]"
  }
  return 0
}

proc nb_netbots {frombot arg} {
  global nb_flag
  set command [lindex [split $arg] 1] ; set bothand [lindex [split $arg] 2]
  if {((![validuser $bothand]) || (![matchattr $bothand b]))} {return 0}
  if {$command == "add"} {
    chattr $bothand +$nb_flag
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: added $bothand to netbot list."
  } elseif {$command == "remove"} {
    chattr $bothand -$nb_flag
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: removed $bothand from netbot list."
  }
  return 0
}

proc nb_dccnetcontrol {hand idx arg} {
  global nb_control nb_group
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netcontrol $arg"
  regsub -all -- ", " $arg "," bots ; regsub -all -- " " $bots "," bots
  if {$bots == ""} {
    if {$nb_control == "*"} {
      putidx $idx "* Controlling all netbots."
    } else {
      regsub -all -- "," $nb_control ", " ctrlbots
      putidx $idx "* Controlling netbots: $ctrlbots"
    }
    if {[info exists nb_group]} {
      putidx $idx "  Available groups: [join [array names nb_group] ", "]"
    }
  } elseif {$bots == "*"} {
    set nb_control "*"
    putidx $idx "* Now controlling all netbots."
  } elseif {[info exists nb_group($bots)]} {
    set nb_control $nb_group($bots)
    regsub -all -- "," $nb_control ", " ctrlbots
    putidx $idx "* Now controlling '$bots' netbots: $ctrlbots"
  } else {
    set nb_control $bots
    regsub -all -- "," $nb_control ", " ctrlbots
    putidx $idx "* Now controlling netbots: $ctrlbots"
  }
  return 0
}

proc nb_dccnetinfo {hand idx arg} {
  global botnet-nick nb_components nb_control nb_ver
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netinfo $arg"
  putidx $idx "BOT        VERSION  COMPONENTS"
  if {[nb_control]} {
    if {[llength $nb_components] > 7} {
      putidx $idx [format "%-10.10s %-8.8s %s" ${botnet-nick} $nb_ver [join [lrange $nb_components 0 6] ", "],]
      if {[llength $nb_components] > 15} {
        putidx $idx [format "%-19s %s" "" [join [lrange $nb_components 7 14] ", "],]
        putidx $idx [format "%-19s %s" "" [join [lrange $nb_components 15 end] ", "]]
      } else {
        putidx $idx [format "%-19s %s" "" [join [lrange $nb_components 7 14] ", "]]
      }
    } else {
      putidx $idx [format "%-10.10s %-8.8s %s" ${botnet-nick} $nb_ver [join [lrange $nb_components 0 6] ", "]]
    }
  }
  nb_sendcmd $nb_control netinfo $idx
  return 0
}

proc nb_netinfo {frombot arg} {
  global nb_components nb_ver
  nb_sendcmd $frombot rnetinfo "[lindex [split $arg] 0] $nb_ver $nb_components"
  return 0
}

proc nb_rnetinfo {frombot arg} {
  set idx [lindex [split $arg] 0] ; set components [lrange [split $arg] 2 end]
  if {[llength $components] > 7} {
    putidx $idx [format "%-10.10s %-8.8s %s" $frombot [lindex [split $arg] 1] [join [lrange $components 0 6] ", "],]
    if {[llength $components] > 15} {
      putidx $idx [format "%-19s %s" "" [join [lrange $components 7 14] ", "],]
      putidx $idx [format "%-19s %s" "" [join [lrange $components 15 end] ", "]]
    } else {
      putidx $idx [format "%-19s %s" "" [join [lrange $components 7 14] ", "]]
    }
  } else {
    putidx $idx [format "%-10.10s %-8.8s %s" $frombot [lindex [split $arg] 1] [join [lrange $components 0 6] ", "]]
  }
  return 0
}

proc nb_dccnetshell {hand idx arg} {
  global botnet-nick nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netshell $arg"
  if {[nb_control]} {
    catch {string trim [exec uptime]} uptime
    if {[info exists uptime]} {
      regsub -all -- "  " $uptime " " uptime
      putidx $idx "${botnet-nick} is on [info hostname] - [unames] - $uptime"
    }
  }
  nb_sendcmd $nb_control netshell "$hand $idx"
  return 0
}

proc nb_netshell {frombot arg} {
  catch {string trim [exec uptime]} uptime
  if {[info exists uptime]} {
    regsub -all -- "  " $uptime " " uptime
    nb_sendcmd $frombot rnetshell "[lindex [split $arg] 1] [info hostname] - [unames] - $uptime"
  }
  return 0
}

proc nb_rnetshell {frombot arg} {
  putidx [lindex [split $arg] 0] "$frombot is on [join [lrange [split $arg] 1 end]]"
  return 0
}

proc nb_dccnetserv {hand idx arg} {
  global botnet-nick nb_control nb_realserver numversion server server-online
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netserv $arg"
  putidx $idx "BOT        SERVER                                CONNECTED"
  if {[nb_control]} {
    catch {set server}
    if {[info exists server]} {
      if {$server == ""} {
        putidx $idx [format "%-10.10s %-37.37s" ${botnet-nick} None]
      } elseif {(([info exists nb_realserver]) && ([string tolower [lindex [split $server :] 0]] != [string tolower $nb_realserver]))} {
        putidx $idx [format "%-10.10s %-37.37s %s" ${botnet-nick} $nb_realserver [nb_duration [expr [unixtime] - ${server-online}]]]
      } else {
        putidx $idx [format "%-10.10s %-37.37s %s" ${botnet-nick} $server [nb_duration [expr [unixtime] - ${server-online}]]]
      }
    }
  }
  nb_sendcmd $nb_control netserv "$hand $idx"
  return 0
}

proc nb_netserv {frombot arg} {
  global nb_realserver server server-online
  catch {set server}
  if {[info exists server]} {
    if {(($server != "") && ([info exists nb_realserver]) && ([string tolower [lindex [split $server :] 0]] != [string tolower $nb_realserver]))} {
      nb_sendcmd $frombot rnetserv "[lindex [split $arg] 1] $nb_realserver [expr [unixtime] - ${server-online}]"
    } else {
      nb_sendcmd $frombot rnetserv "[lindex [split $arg] 1] $server [expr [unixtime] - ${server-online}]"
    }
  }
  return 0
}

proc nb_rnetserv {frombot arg} {
  global numversion
  set server [lindex [split $arg] 1]
  if {$server != ""} {
    putidx [lindex [split $arg] 0] [format "%-10.10s %-37.37s %s" $frombot $server [nb_duration [lindex [split $arg] 2]]]
  } else {
    putidx [lindex [split $arg] 0] [format "%-10.10s %-37.37s %s" $frombot none n/a]
  }
  return 0
}

proc nb_dccnetpass {hand idx arg} {
  global botnet-nick nb_control nb_flag nb_key
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netpass"
  set thepass [lindex [split $arg] 0] ; set forhand [lindex [split $arg] 1]
  if {$forhand == ""} {
    if {$thepass == "-check"} {
      putidx $idx "Checking netbots for users without passwords.."
      if {[nb_control]} {
        set nopass ""
        foreach user [userlist] {
          if {[passwdok $user ""]} {
            lappend nopass $user
          }
        }
        if {$nopass != ""} {
          putidx $idx "${botnet-nick}: [join $nopass ", "]"
        } else {
          putidx $idx "${botnet-nick}: all users have a password set."
        }
      }
      nb_sendcmd $nb_control netpass "$hand $idx [encrypt $nb_key $thepass]"
    } else {
      if {(($nb_flag == "b") || ($nb_flag == "all"))} {
        putidx $idx "You cannot change the netbot password if nb_flag is set to 'b' or 'all'."
        return 0
      }
      if {$thepass == ""} {
        set thepass [nb_randpass]
      }
      foreach bot [userlist $nb_flag] {
        setuser $bot PASS $thepass
      }
      nb_sendcmd * netpass "$hand $idx [encrypt $nb_key $thepass]"
      putidx $idx "Changed netbot password to '$thepass'."
    }
  } elseif {(($thepass != "") && ($forhand != ""))} {
    if {![validuser $forhand]} {
      putidx $idx "$forhand is not a valid user." ; return 0
    }
    if {[nb_control]} {
      setuser $forhand PASS $thepass
    }
    nb_sendcmd $nb_control netpass "$hand $idx [encrypt $nb_key $thepass] $forhand"
    putidx $idx "Changed password for $forhand to '$thepass'."
  }
  return 0
}

proc nb_netpass {frombot arg} {
  global nb_flag nb_key
  set thepass [decrypt $nb_key [lindex [split $arg] 2]] ; set forhand [lindex [split $arg] 3]
  if {$forhand == ""} {
    if {$thepass == "-check"} {
      set nopass ""
      foreach user [userlist] {
        if {[passwdok $user ""]} {
          lappend nopass $user
        }
      }
      if {$nopass != ""} {
        nb_sendcmd $frombot rnetpass "[lindex [split $arg] 1] $nopass"
      } else {
        nb_sendcmd $frombot rnetpass [lindex [split $arg] 1]
      }
    } else {
      foreach bot [userlist $nb_flag] {
        setuser $bot PASS $thepass
      }
      putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: changed netbot password."
    }
  } else {
    setuser $forhand PASS $thepass
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: changed password for $forhand."
  }
  return 0
}

proc nb_rnetpass {frombot arg} {
  set nopass [lrange [split $arg] 1 end]
  if {$nopass != ""} {
    putidx [lindex [split $arg] 0] "$frombot: [join $nopass ", "]"
  } else {
    putidx [lindex [split $arg] 0] "$frombot: all users have a password set."
  }
  return 0
}

proc nb_dccnetsave {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netsave $arg"
  putidx $idx "Performing netbot userfile/chanfile save."
  if {[nb_control]} {
    save
  }
  nb_sendcmd $nb_control netsave $hand
  return 0
}

proc nb_netsave {frombot arg} {
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: performing netbot userfile/chanfile save."
  save
  return 0
}

proc nb_dccnethash {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# nethash $arg"
  putidx $idx "Rehashing netbots."
  if {[nb_control]} {
    uplevel {rehash}
  }
  nb_sendcmd $nb_control nethash $hand
  return 0
}

proc nb_nethash {frombot arg} {
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: rehashing."
  uplevel {rehash}
  return 0
}

proc nb_dccnetsay {hand idx arg} {
  global botnet-nick nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netsay $arg"
  set sayto [lindex [split $arg] 0] ; set themsg [join [lrange [split $arg] 1 end]]
  if {(($sayto == "") || ($themsg == ""))} {
    putidx $idx "Usage: netsay <channel|nick> <message>" ; return 0
  }
  nb_sendcmd $nb_control netsay "$hand $sayto $themsg"
  if {[nb_control]} {
    if {[info commands putserv] != ""} {
      putserv "PRIVMSG $sayto :$themsg"
    }
  }
  return 0
}

proc nb_netsay {frombot arg} {
  set sayto [lindex [split $arg] 1]
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: sending message to $sayto"
  if {[info commands putserv] != ""} {
    putserv "PRIVMSG $sayto :[join [lrange [split $arg] 2 end]]"
  }
  return 0
}

proc nb_dccnetact {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netact $arg"
  set actto [lindex [split $arg] 0] ; set theact [join [lrange [split $arg] 1 end]]
  if {(($actto == "") || ($theact == ""))} {
    putidx $idx "Usage: netact <channel|nick> <action>" ; return 0
  }
  nb_sendcmd $nb_control netact "$hand $actto $theact"
  if {[nb_control]} {
    if {[info commands putserv] != ""} {
      putserv "PRIVMSG $actto :\001ACTION $theact\001"
    }
  }
  return 0
}

proc nb_netact {frombot arg} {
  set actto [lindex [split $arg] 1]
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: sending action to $actto"
  if {[info commands putserv] != ""} {
    putserv "PRIVMSG $actto :\001ACTION [join [lrange [split $arg] 2 end]]\001"
  }
  return 0
}

proc nb_dccnetnotc {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netnotice $arg"
  set notcto [lindex [split $arg] 0] ; set thenotc [join [lrange [split $arg] 1 end]]
  if {(($notcto == "") || ($thenotc == ""))} {
    putidx $idx "Usage: netnotice <channel|nick> <notice>" ; return 0
  }
  nb_sendcmd $nb_control netnotc "$hand $notcto $thenotc"
  if {[nb_control]} {
    if {[info commands putserv] != ""} {
      putserv "NOTICE $notcto :$thenotc"
    }
  }
  return 0
}

proc nb_netnotc {frombot arg} {
  set notcto [lindex [split $arg] 1]
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: sending notice to $notcto"
  if {[info commands putserv] != ""} {
    putserv "NOTICE $notcto :[join [lrange [split $arg] 2 end]]"
  }
  return 0
}

proc nb_dccnetdump {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netdump $arg"
  if {$arg == ""} {
    putidx $idx "Usage: netdump <text>" ; return 0
  }
  putidx $idx "Making netbots dump text '$arg'."
  nb_sendcmd $nb_control netdump "$hand $arg"
  if {[nb_control]} {
    if {[info commands putserv] != ""} {
      putserv $arg
    }
  }
  return 0
}

proc nb_netdump {frombot arg} {
  set text [join [lrange [split $arg] 1 end]]
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: dumping text '$text'."
  if {[info commands putserv] != ""} {
    putserv $text
  }
  return 0
}

proc nb_dccnetnick {hand idx arg} {
  global altnick botnet-nick nb_control nick
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netnick $arg"
  if {$arg == ""} {
    putidx $idx "Changing netbot nicks to altnicks."
    nb_sendcmd $nb_control netnick $hand
    if {[nb_control]} {
      set newnick $altnick
      while {[regsub -- \\? $newnick [rand 10] newnick]} {continue}
      set nick $newnick
    }
  } elseif {$arg == "-rand"} {
    putidx $idx "Changing netbot nicks to randnicks."
    nb_sendcmd $nb_control netnick "$hand -rand"
    if {[nb_control]} {
      set nick [string range $nick 0 2][expr 10 + [rand 999990]]
    }
  } else {
    putidx $idx "Changing netbot nicks."
    nb_sendcmd $nb_control netnick "$hand $arg"
    if {[nb_control]} {
      foreach i [split $arg] {
        if {(([string tolower [lindex [split $i :] 0]] == [string tolower ${botnet-nick}]) && ([set newnick [lindex [split $i :] 1]] != ""))} {
          set nick $newnick
        }
      }
    }
  }
  return 0
}

proc nb_netnick {frombot arg} {
  global altnick botnet-nick nick
  if {[lindex [split $arg] 1] == ""} {
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: changing nick to altnick '$altnick'."
    set newnick $altnick
    while {[regsub -- \\? $newnick [rand 10] newnick]} {continue}
    set nick $newnick
  } elseif {[lindex [split $arg] 1] == "-rand"} {
    set newnick [string range $nick 0 2][expr 10 + [rand 999990]]
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: changing nick to randnick '$newnick'."
    set nick $newnick
  } else {
    foreach i [lrange [split $arg] 1 end] {
      if {(([string tolower [lindex [split $i :] 0]] == [string tolower ${botnet-nick}]) && ([set newnick [lindex [split $i :] 1]] != ""))} {
        putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: changing nick to '$newnick'."
        set nick $newnick
      }
    }
  }
  return 0
}

proc nb_dccnetjump {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netjump $arg"
  if {[lindex [split $arg] 0] == ""} {
    putidx $idx "Usage: netjump <bot|-all> \[server\]" ; return 0
  }
  set bot [lindex [split $arg] 0] ; set server [lindex [split $arg] 1]
  if {$bot == "-all"} {
    nb_sendcmd $nb_control netjump "$hand $server"
    if {$server == ""} {
      putidx $idx "Making netbots jump to next server."
      if {[nb_control]} {
        jump
      }
    } else {
      putidx $idx "Making netbots jump to $server"
      if {[nb_control]} {
        eval jump [join [split $server :]]
      }
    }
  } else {
    nb_sendcmd $bot netjump "$hand $server"
    if {$server == ""} {
      putidx $idx "Making $bot jump to next server."
    } else {
      putidx $idx "Making $bot jump to $server"
    }
  }
  return 0
}

proc nb_netjump {frombot arg} {
  set server [lindex [split $arg] 1]
  if {$server == ""} {
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: jumping to next server."
    jump
  } else {
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: jumping to $server"
    eval jump [join [split $server :]]
  }
  return 0
}

proc nb_dccnetjoin {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netjoin $arg"
  set chan [lindex [split $arg] 0] ; set key [lindex [split $arg] 1]
  if {(($chan == "") || ([string trim [string index $chan 0] "#+&!"] != ""))} {
    putidx $idx "Usage: netjoin <channel>" ; return 0
  }
  putidx $idx "Adding channel $chan to netbots."
  nb_sendcmd $nb_control netjoin "$hand $chan $key"
  if {[nb_control]} {
    if {![validchan $chan]} {
      channel add $chan
      nb_chanset $chan
      if {[info commands bop_setneed] != ""} {
        bop_setneed $chan
      }
      if {$key != ""} {
        putserv "JOIN $chan $key"
      }
    }
  }
  return 0
}

proc nb_netjoin {frombot arg} {
  set chan [lindex [split $arg] 1]
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: adding channel $chan"
  if {![validchan $chan]} {
    channel add $chan
    nb_chanset $chan
    if {[info commands bop_setneed] != ""} {
      bop_setneed $chan
    }
    if {[set key [lindex [split $arg] 2]] != ""} {
      putserv "JOIN $chan $key"
    }
  }
  return 0
}

proc nb_dccnetpart {hand idx arg} {
  global nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netpart $arg"
  set chan [lindex [split $arg] 0]
  if {(($chan == "") || ([string trim [string index $chan 0] "#+&!"] != ""))} {
    putidx $idx "Usage: netpart <channel>" ; return 0
  }
  putidx $idx "Removing channel $chan from netbots."
  nb_sendcmd $nb_control netpart "$hand $chan"
  if {[nb_control]} {
    if {(([validchan $chan]) && ([isdynamic $chan]))} {
      channel remove $chan
    }
  }
  return 0
}

proc nb_netpart {frombot arg} {
  set chan [lindex [split $arg] 1]
  putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: removing channel $chan"
  if {(([validchan $chan]) && ([isdynamic $chan]))} {
    channel remove $chan
  }
  return 0
}

proc nb_dccnetchanset {hand idx arg} {
  global config nb_control
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netchanset $arg"
  set chan [lindex [split $arg] 0]
  if {$chan == "-reload"} {
    putidx $idx "Making netbots reload channel settings from config file."
    nb_sendcmd $nb_control netchanset "$hand $chan"
    if {![file exists $config]} {
      putidx $idx "Error: unable to find file $config" ; return 0
    }
    set rfile [open $config r]
    while {![eof $rfile]} {
      set string [gets $rfile]
      if {[string match "channel set *" $string]} {
        eval $string
      }
    }
    close $rfile
    save
    return 0
  }
  if {(($chan == "") || (([string trim [string index $chan 0] "#+&!"] != "") && ($chan != "-all")))} {
    putidx $idx "Usage: netchanset <channel> \[settings\]" ; return 0
  }
  regsub -all -- \" [join [lrange [split $arg] 1 end]] "" settings
  set settings [split $settings]
  if {$settings == ""} {
    if {$chan == "-all"} {
      putidx $idx "Setting channel settings for all channels to netbots default settings."
    } else {
      putidx $idx "Setting channel settings for $chan to netbots default settings."
    }
    nb_sendcmd $nb_control netchanset "$hand $chan"
    if {[nb_control]} {
      if {$chan == "-all"} {
        foreach chan [channels] {
          nb_chanset $chan
        }
      } else {
        if {[validchan $chan]} {
          nb_chanset $chan
        }
      }
    }
  } else {
    if {[string match "need-*" [lindex $settings 0]]} {
      putidx $idx "You cannot change need-op/invite/key/limit/unban settings." ; return 0
    }
    if {$chan == "-all"} {
      putidx $idx "Setting channel settings for all channels to '$settings' on netbots."
    } else {
      putidx $idx "Setting channel settings for $chan to '$settings' on netbots."
    }
    nb_sendcmd $nb_control netchanset "$hand $chan $settings"
    if {[nb_control]} {
      if {$chan == "-all"} {
        if {(([string index [lindex $settings 0] 0] == "+") || ([string index [lindex $settings 0] 0] == "-"))} {
          foreach chan [channels] {
            eval channel set $chan $settings
          }
        } else {
          foreach chan [channels] {
            channel set $chan [lindex $settings 0] [lrange $settings 1 end]
          }
        }
      } else {
        if {[validchan $chan]} {
          if {(([string index [lindex $settings 0] 0] == "+") || ([string index [lindex $settings 0] 0] == "-"))} {
            eval channel set $chan $settings
          } else {
            channel set $chan [lindex $settings 0] [lrange $settings 1 end]
          }
        }
      }
    }
  }
  return 0
}

proc nb_netchanset {frombot arg} {
  global config
  set chan [lindex [split $arg] 1]
  if {$chan == "-reload"} {
    putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: reloading channel settings from config file."
    if {![file exists $config]} {
      putlog "netbots: error - unable to find file $config" ; return 0
    }
    set rfile [open $config r]
    while {![eof $rfile]} {
      set string [gets $rfile]
      if {[string match "channel set *" $string]} {
        eval $string
      }
    }
    close $rfile
    save
    return 0
  }
  set settings [lrange [split $arg] 2 end]
  if {$settings == ""} {
    if {$chan == "-all"} {
      putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: setting channel settings for all channels to netbots default settings."
      foreach chan [channels] {
        nb_chanset $chan
      }
    } else {
      putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: setting channel settings for $chan to netbots default settings."
      if {[validchan $chan]} {
        nb_chanset $chan
      }
    }
  } else {
    if {[string match "need-*" [lindex $settings 0]]} {return 0}
    if {$chan == "-all"} {
      putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: setting channel settings for all channels to '$settings'."
      if {(([string index [lindex $settings 0] 0] == "+") || ([string index [lindex $settings 0] 0] == "-"))} {
        foreach chan [channels] {
          eval channel set $chan $settings
        }
      } else {
        foreach chan [channels] {
          channel set $chan [lindex $settings 0] [lrange $settings 1 end]
        }
      }
    } else {
      putlog "netbots \[[lindex [split $arg] 0]@$frombot\]: setting channel settings for $chan to '$settings'."
      if {[validchan $chan]} {
        if {(([string index [lindex $settings 0] 0] == "+") || ([string index [lindex $settings 0] 0] == "-"))} {
          eval channel set $chan $settings
        } else {
          channel set $chan [lindex $settings 0] [lrange $settings 1 end]
        }
      }
    }
  }
  return 0
}

proc nb_dccnetupdate {hand idx arg} {
  global nb_control nb_flag nb_sending
  if {![nb_checkowner $hand $idx]} {return 0}
  putcmdlog "#$hand# netupdate $arg"
  set command [lindex [split $arg] 0]
  switch -exact -- $command {
    "-config" {set file "config"}
    "-motd" {set file "motd"}
    "-tbanner" {set file "telnet-banner"}
    "-settings" {set file "settings"}
    "-scripts" {set file "scripts"}
    "-everything" {set file "everything"}
    "-file" {set file [lindex [split $arg] 1]}
    "default" {set file "settings/scripts"}
  }
  if {[string match -* $command]} {
    if {$file == "settings/scripts"} {
      putidx $idx "Valid options are: -config, -motd, -tbanner, -settings, -scripts, -everything, -file" ; return 0
    }
    if {$command != "-file"} {
      set bot [lindex [split $arg] 1]
    } else {
      if {$file == ""} {
        putidx $idx "You must specify a file." ; return 0
      }
      if {![file exists $file]} {
        putidx $idx "File '$file' does not exist." ; return 0
      }
      if {[file isdirectory $file]} {
        putidx $idx "File '$file' is a directory." ; return 0
      }
      set bot [lindex [split $arg] 2]
    }
  } else {
    set bot [lindex [split $arg] 0]
  }
  if {[isbotnetnick $bot]} {
    putidx $idx "Cannot send files to myself." ; return 0
  }
  if {$bot == ""} {
    if {[bots] == ""} {
      putidx $idx "There are no linked bots." ; return 0
    }
    putidx $idx "Initiating transfer of $file to netbots.."
    nb_sendcmd $nb_control netupdate "initiate $file $hand"
  } else {
    if {(($nb_flag != "all") && (![matchattr $bot $nb_flag]))} {
      putidx $idx "$bot is not a valid netbot." ; return 0
    }
    if {![islinked $bot]} {
      putidx $idx "$bot is not on the botnet." ; return 0
    }
    if {[info exists nb_sending([string tolower $bot])]} {
      putidx $idx "File transfer to $bot is already active." ; return 0
    }
    putidx $idx "Initiating transfer of $file to $bot.."
    nb_sendcmd $bot netupdate "initiate $file $hand"
  }
  return 0
}

proc nb_netupdate {frombot arg} {
  global config motd nb_dir nb_files nb_getting nb_lines nb_scripts nb_setscript nb_timeout telnet-banner
  switch -exact -- [lindex [split $arg] 0] {
    "initiate" {
      if {$nb_getting} {return 0}
      set file [lindex [split $arg] 1] ; set hand [lindex [split $arg] 2]
      if {$hand != ""} {
        putlog "netbots \[$hand@$frombot\]: initiating $file update.."
      } else {
        putlog "netbots \[$frombot\]: initiating $file update.."
      }
      switch -exact -- $file {
        "config" {nb_sendcmd $frombot rnetupdate "send $file $config"}
        "motd" {nb_sendcmd $frombot rnetupdate "send $file $motd"}
        "telnet-banner" {
          if {[info exists telnet-banner]} {
            nb_sendcmd $frombot rnetupdate "send $file ${telnet-banner}"
          }
        }
        "settings" {nb_sendcmd $frombot rnetupdate "send $file $nb_setscript"}
        "scripts" {nb_sendcmd $frombot rnetupdate "send $file $nb_scripts"}
        "settings/scripts" {nb_sendcmd $frombot rnetupdate "send $file $nb_setscript $nb_scripts"}
        "everything" {
          if {[info exists telnet-banner]} {
            nb_sendcmd $frombot rnetupdate "send $file $nb_setscript $nb_scripts $config $motd ${telnet-banner}"
          } else {
            nb_sendcmd $frombot rnetupdate "send $file $nb_setscript $nb_scripts $config $motd"
          }
        }
        "default" {nb_sendcmd $frombot rnetupdate "send $file $file"}
      }
    }
    "wait" {
      if {$nb_getting} {return 0}
      set file [lindex [split $arg] 1]
      putlog "netbots: $frombot is busy, waiting to try again.."
      set nb_getting 1
      switch -exact -- $file {
        "scripts" {
          utimer [expr 20 + [rand 21]] [split "nb_wait $frombot $file"]
        }
        "settings/scripts" {
          utimer [expr 20 + [rand 21]] [split "nb_wait $frombot $file"]
        }
        "everything" {
          utimer [expr 20 + [rand 31]] [split "nb_wait $frombot $file"]
        }
        "default" {
          utimer [expr 10 + [rand 11]] [split "nb_wait $frombot $file"]
        }
      }
    }
    "start" {
      if {$nb_getting} {return 0}
      set file [lindex [split $arg] 1]
      set nb_getting 1
      putlog "netbots: downloading $file from $frombot"
      utimer $nb_timeout [split "nb_timeout getting $frombot $file"]
      set nb_lines ""
    }
    "write" {
      set nb_lines [string range $arg [expr [string first " " $arg] + 1] end]
    }
    "close" {
      if {!$nb_getting} {return 0}
      set filename [lindex [split $arg] 1]
      lappend nb_files [list $filename [file tail $filename].tmp [lindex [split $arg] 2] $nb_lines]
      set nb_lines ""
    }
    "end" {
      if {!$nb_getting} {return 0}
      set file [lindex [split $arg] 1]
      foreach filepack $nb_files {
        set tmpfile [lindex $filepack 1]
        lappend tmplist $tmpfile
        set wfile [open $tmpfile w]
        fconfigure $wfile -translation binary
        puts -nonewline $wfile [lindex $filepack 3]
        close $wfile
      }
      foreach filepack $nb_files {
        set filename [lindex $filepack 0]
        set chksize [lindex $filepack 2]
        if {[catch {file size [lindex $filepack 1]} filesize]} {
          set filesize 0
        }
        if {(($filesize != $chksize) || ($chksize == 0))} {
          putlog "netbots: error on $file download from $frombot - filesize mismatch for $filename (expected $chksize bytes, got $filesize [nb_plural byte $filesize])."
          nb_sendcmd $frombot rnetupdate "fserror $file"
          nb_killutimer "nb_timeout getting $frombot *"
          eval file delete -- [join $tmplist]
          unset nb_files
          set nb_getting 0
          return 0
        }
        if {[set dir [file dirname $filename]] != "."} {
          if {![file exists $dir]} {
            putlog "netbots: creating directory '$dir'.."
            file mkdir $dir
          } elseif {![file isdirectory $dir]} {
            putlog "netbots: error on $file download from $frombot - '$dir' exists but is not a directory."
            nb_sendcmd $frombot rnetupdate "direrror $file"
            nb_killutimer "nb_timeout getting $frombot *"
            eval file delete -- [join $tmplist]
            unset nb_files
            set nb_getting 0
            return 0
          }
        }
      }
      foreach filepack $nb_files {
        set filename [lindex $filepack 0]
        file rename -force -- [lindex $filepack 1] $filename
        if {$filename == $config} {
          catch {exec chmod 700 $filename}
        } else {
          catch {exec chmod 600 $filename}
        }
      }
      nb_killutimer "nb_timeout getting $frombot *"
      unset nb_files
      nb_sendcmd $frombot rnetupdate "end $file"
      if {(($file == "motd") || ($file == "telnet-banner"))} {
        putlog "netbots: $file download complete"
        set nb_getting 0
      } else {
        putlog "netbots: $file download complete, rehashing.."
        utimer 3 rehash
        utimer 5 {set nb_getting 0}
      }
    }
    "ferror" {
      putlog "netbots: error - $frombot cannot find file '[lindex [split $arg] 1]'."
    }
  }
  return 0
}

proc nb_rnetupdate {frombot arg} {
  global nb_max nb_sending nb_timeout
  set file [lindex [split $arg] 1] ; set filenames [lrange [split $arg] 2 end]
  switch -exact -- [lindex [split $arg] 0] {
    "initiate" {
      if {[info exists nb_sending([string tolower $frombot])]} {return 0}
      nb_sendcmd $frombot netupdate "initiate $file"
    }
    "send" {
      set stlfrombot [string tolower $frombot]
      if {[info exists nb_sending($stlfrombot)]} {return 0}
      if {[llength [array names nb_sending]] >= $nb_max} {
        nb_sendcmd $frombot netupdate "wait $file"
        return 0
      }
      foreach filename $filenames {
        if {![file exists $filename]} {
          putlog "netbots: error - cannot find file '$filename' (requested by $frombot)."
          nb_sendcmd $frombot netupdate "ferror $filename"
          return 0
        }
      }
      putlog "netbots: sending $file to $frombot"
      set nb_sending($stlfrombot) 1
      utimer $nb_timeout [split "nb_timeout sending $frombot $file"]
      nb_sendcmd $frombot netupdate "start $file"
      foreach filename $filenames {
        set rfile [open $filename r]
        fconfigure $rfile -translation binary
        set fstring [read $rfile]
        close $rfile
        nb_sendcmd $frombot netupdate "write $fstring"
        if {[catch {file size $filename} filesize]} {
          set filesize 0
        }
        nb_sendcmd $frombot netupdate "close $filename $filesize"
      }
      nb_sendcmd $frombot netupdate "end $file"
    }
    "end" {
      set stlfrombot [string tolower $frombot]
      nb_killutimer "nb_timeout sending $frombot *"
      if {[info exists nb_sending($stlfrombot)]} {
        unset nb_sending($stlfrombot)
      }
      putlog "netbots: completed $file transfer to $frombot"
    }
    "fserror" {
      set stlfrombot [string tolower $frombot]
      nb_killutimer "nb_timeout sending $frombot *"
      if {[info exists nb_sending($stlfrombot)]} {
        unset nb_sending($stlfrombot)
      }
      putlog "netbots: error on $file transfer to $frombot - filesize mismatch."
    }
    "direrror" {
      set stlfrombot [string tolower $frombot]
      nb_killutimer "nb_timeout sending $frombot *"
      if {[info exists nb_sending($stlfrombot)]} {
        unset nb_sending($stlfrombot)
      }
      putlog "netbots: error on $file transfer to $frombot - $frombot is unable to create directory."
    }
  }
  return 0
}

proc nb_wait {frombot file} {
  global config motd nb_getting nb_scripts nb_setscript telnet-banner
  set nb_getting 0
  switch -exact -- $file {
    "config" {nb_sendcmd $frombot rnetupdate "initiate $file $config"}
    "motd" {nb_sendcmd $frombot rnetupdate "initiate $file $motd"}
    "telnet-banner" {
      if {[info exists telnet-banner]} {
        nb_sendcmd $frombot rnetupdate "initiate $file ${telnet-banner}"
      }
    }
    "settings" {nb_sendcmd $frombot rnetupdate "initiate $file $nb_setscript"}
    "scripts" {nb_sendcmd $frombot rnetupdate "initiate $file $nb_scripts"}
    "settings/scripts" {nb_sendcmd $frombot rnetupdate "initiate $file $nb_setscript $nb_scripts"}
    "everything" {
      if {[info exists telnet-banner]} {
        nb_sendcmd $frombot rnetupdate "initiate $file $nb_setscript $nb_scripts $config $motd ${telnet-banner}"
      } else {
        nb_sendcmd $frombot rnetupdate "initiate $file $nb_setscript $nb_scripts $config $motd"
      }
    }
    "default" {nb_sendcmd $frombot rnetupdate "initiate $file $file"}
  }
  return 0
}

proc nb_timeout {action bot file} {
  global nb_files nb_getting nb_lines nb_sending
  switch -exact -- $action {
    "getting" {
      putlog "netbots: timed out $file download from $bot"
      if {[info exists nb_lines]} {
        unset nb_lines
      }
      if {[info exists nb_files]} {
        unset nb_files
      }
      set nb_getting 0
    }
    "sending" {
      putlog "netbots: timed out $file transfer to $bot"
      if {[info exists nb_sending([string tolower $bot])]} {
        unset nb_sending([string tolower $bot])
      }
    }
  }
  return 0
}

proc nb_tautopass {min hour day month year} {
  global nb_flag nb_key
  putlog "netbots: initiating netbot password change.."
  set botlist ""
  foreach bot [userlist $nb_flag] {
    if {((![isbotnetnick $bot]) && (![islinked $bot]))} {
      lappend botlist $bot
    }
  }
  if {$botlist != ""} {
    set botlist [join $botlist ", "]
    putlog "netbots: cannot change netbot password because not all netbots are linked (currently missing: $botlist)."
    return 0
  }
  set thepass [nb_randpass]
  foreach bot [userlist $nb_flag] {
    setuser $bot PASS $thepass
  }
  save
  nb_sendcmd * autopass [encrypt $nb_key $thepass]
  putlog "netbots: changed netbot password."
  return 0
}

proc nb_autopass {frombot arg} {
  global nb_flag nb_key
  set thepass [decrypt $nb_key [lindex [split $arg] 0]]
  foreach bot [userlist $nb_flag] {
    setuser $bot PASS $thepass
  }
  save
  putlog "netbots \[$frombot\]: changed netbot password."
  return 0
}

proc nb_chanset {chan} {
  global nb_chanmode nb_chansets nb_floodchan nb_flooddeop nb_floodkick nb_floodjoin nb_floodctcp nb_idlekick
  channel set $chan chanmode $nb_chanmode
  channel set $chan idle-kick $nb_idlekick
  channel set $chan flood-chan $nb_floodchan
  channel set $chan flood-deop $nb_flooddeop
  channel set $chan flood-kick $nb_floodkick
  channel set $chan flood-join $nb_floodjoin
  channel set $chan flood-ctcp $nb_floodctcp
  foreach chanset $nb_chansets {
    channel set $chan $chanset
  }
  return 0
}

proc nb_update {min hour day month year} {
  global nb_dir nb_numver nb_update
  timer 3 nb_http
  if {$nb_update != ""} {
    if {[info commands sendnote] == ""} {
      putlog "netbots: error - update check cannot find the 'sendnote' command (the nb_update feature requires the notes module)." ; return 0
    }
    if {[catch {package require http}]} {
      putlog "netbots: error - update check cannot find package 'http' (the nb_update feature requires tcl 8.0 or later)." ; return 0
    }
    if {![file exists $nb_dir/nbupdate]} {
      set file [open $nb_dir/nbupdate w] ; puts $file $nb_numver ; close $file ; unset file
    }
    set ufile [open $nb_dir/nbupdate r] ; set unumver [lindex [split [gets $ufile]] 0] ; close $ufile ; unset ufile
    if {[catch {::http::geturl http://www.egghelp.org/netbots/nbupdate.txt} token]} {
      putlog "netbots: error - update check is unable to connect to server (must be down or something)." ; return 0
    }
    set hdata [::http::data $token]
    unset $token
    if {[lindex [split $hdata] 0] != "nbupdate"} {
      putlog "netbots: error - update check did not receive expected response from server." ; return 0
    }
    set hnumver [lindex [split $hdata] 1]
    set ufile [open $nb_dir/nbupdate w] ; puts $ufile $hnumver ; close $ufile ; unset ufile
    if {$hnumver > $unumver} {
      foreach recipient [split $nb_update] {
        if {[validuser $recipient]} {
          sendnote NETBOTS $recipient [join [lrange [split $hdata] 2 end]]
        }
      }
    }
  }
  return 0
}

proc nb_http {} {
  global nb_numver
  if {![catch {socket www.egghelp.org 80} http]} {
    fconfigure $http -translation {auto crlf}
    puts $http "GET /netbots/nb$nb_numver.txt HTTP/1.0"
    puts $http "Accept: */*"
    puts $http "Host: www.egghelp.org"
    puts $http "User-Agent: netbots.tcl http client"
    puts $http ""
    utimer 1 [split "close $http"]
  }
  return 0
}

proc nb_chon {hand idx} {
  global nb_chon nb_control nb_defctrl
  set nb_control $nb_defctrl
  global nb_flag nb_ver
  if {$nb_chon} {
    if {[matchattr $hand mnt]} {
      set bots [expr [llength [bots]] + 1]
      putidx $idx " "
      putidx $idx "Running netbots.tcl $nb_ver"
      putidx $idx "  $bots [nb_plural bot $bots] on the botnet."
      if {(($nb_flag != "b") && ($nb_flag != "all"))} {
        set botlist "" ; set offline ""
        foreach netbot [userlist $nb_flag] {
          if {(([islinked $netbot]) || ([isbotnetnick $netbot]))} {
            lappend botlist $netbot
          } else {
            lappend offline $netbot
          }
        }
        if {$botlist == ""} {
          putidx $idx "  No netbots linked."
        } else {
          set bots [llength $botlist]
          if {$offline != ""} {
            putidx $idx "  $bots [nb_plural netbot $bots] ([llength $offline] offline or not linked: [join $offline ", "])."
          } else {
            putidx $idx "  $bots [nb_plural netbot $bots] (all netbots online and linked)."
          }
        }
      }
    }
    set partylist ""
    putidx $idx " "
    foreach partydude [whom *] {
      if {[lindex $partydude 0] == $hand} {continue}
      lappend partylist [lindex $partydude 0]@[lindex $partydude 1]
    }
    if {$partylist == ""} {
      putidx $idx "There are no other users on the bot or party line."
      putidx $idx " "
    } else {
      putidx $idx "Other users on the bot and party line: [join $partylist ", "]"
      putidx $idx " "
    }
    nb_sendcmd * rbroadcast "$hand has logged in."
  }
  return 0
}

proc nb_chof {hand idx} {
  nb_sendcmd * rbroadcast "$hand has logged out."
  return 0
}

proc nb_sighup {type} {
  nb_sendcmd * rbroadcast "Received SIGHUP!" ; return 0
}

proc nb_sigterm {type} {
  nb_sendcmd * rbroadcast "Received SIGTERM!" ; return 0
}

proc nb_svkline {from keyword arg} {
  nb_sendcmd * rbroadcast "KLINED from $from: [string range [string trimleft $arg :] 0 150]" ; return 0
}

proc nb_svconnect {from keyword arg} {
  global nb_realserver
  set nb_realserver $from
  if {$nb_realserver == ""} {
    unset nb_realserver
  }
  nb_sendcmd * rbroadcast "Connected to $from" ; return 0
}

proc nb_cmdcast {idx text} {
  set hand [idx2hand $idx] ; set cmd [string tolower $text]
  if {![catch {getchan $idx} chan]} {
    if {$chan == -1} {
      set cmd .[string trimleft $cmd "."]
    } elseif {![string match ".*" $cmd]} {
      return $text
    }
    if {(([string match ".chpass *" $cmd]) || ([string match ".newpass *" $cmd]))} {
      nb_sendcmd * rbroadcast "#$hand# [string trimleft [lindex [split $text] 0] "."] .."
    } elseif {((![string match ".note *" $cmd]) && (![string match ".????note *" $cmd]))} {
      nb_sendcmd * rbroadcast "#$hand# [string trimleft [string range $text 0 150] "."]"
    }
  }
  return $text
}

proc nb_broadcast {frombot arg} {
  global nb_broadcast
  if {$nb_broadcast == ""} {return 0}
  foreach dccuser [dcclist] {
    if {[lindex $dccuser 3] == "CHAT"} {
      set idx [lindex $dccuser 0]
      if {[matchattr [idx2hand $idx] $nb_broadcast]} {
        putidx $idx "\[$frombot\]: $arg"
      }
    }
  }
  return 0
}

proc nb_control {} {
  global botnet-nick nb_control
  if {(($nb_control == "*") || ([lsearch -exact [split [string tolower $nb_control] ,] [string tolower ${botnet-nick}]] != -1))} {return 1}
  return 0
}

proc nb_sendcmd {to cmd arg} {
  global botnet-nick nb_flag nb_key
  set key [encrypt $nb_key [list ${botnet-nick} $cmd $arg]]
  while {$key != ""} {
    lappend skey [string range $key 0 390]
    set key [string range $key 391 end]
  }
  if {$to == "*"} {
    if {$nb_flag == "all"} {
      if {[llength $skey] > 1} {
        set sid [expr [rand 999] + 1]
        foreach key $skey {
          putallbots "nb [list $sid $key]"
        }
        putallbots "nb $sid"
      } else {
        putallbots "nb [list 0 $skey]"
      }
    } else {
      foreach bot [userlist $nb_flag] {
        if {[islinked $bot]} {
          if {[llength $skey] > 1} {
            set sid [expr [rand 999] + 1]
            foreach key $skey {
              putbot $bot "nb [list $sid $key]"
            }
            putbot $bot "nb $sid"
          } else {
            putbot $bot "nb [list 0 $skey]"
          }
        }
      }
    }
  } else {
    foreach bot [split $to ,] {
      if {[islinked $bot]} {
        if {[llength $skey] > 1} {
          set sid [expr [rand 999] + 1]
          foreach key $skey {
            putbot $bot "nb [list $sid $key]"
          }
          putbot $bot "nb $sid"
        } else {
          putbot $bot "nb [list 0 $skey]"
        }
      }
    }
  }
  return 0
}

proc nb_gotcmd {frombot cmd arg} {
  global nb_ctrlbot nb_ctrlbots nb_flag nb_gbuffer nb_key nb_netcmds nb_nudisable
  if {(($nb_flag != "all") && (![matchattr $frombot $nb_flag]))} {return 0}
  if {[set sid [lindex $arg 0]]} {
    set skey [lindex $arg 1]
    if {$skey != ""} {
      append nb_gbuffer($sid) $skey
      return 0
    }
    set arg [decrypt $nb_key [append nb_gbuffer($sid) $skey]]
    unset nb_gbuffer($sid)
  } else {
    set arg [decrypt $nb_key [lindex $arg 1]]
  }
  if {[lindex $arg 0] != $frombot} {
    putlog "netbots: rejected command from $frombot (incorrect key)." ; return 0
  }
  set command [lindex $arg 1]
  if {![string match "r*" $command]} {
    if {[array exists nb_ctrlbots]} {
      if {[lsearch -exact $nb_ctrlbot(*) $command] == -1} {
        set stlfrombot [string tolower $frombot]
        if {((![info exists nb_ctrlbot($stlfrombot)]) || (($nb_ctrlbot($stlfrombot) != "*") && ([lsearch -exact $nb_ctrlbot($stlfrombot) $command] == -1)))} {
          putlog "netbots: rejected command from $frombot (not a control bot)." ; return 0
        }
      }
    } else {
      if {(($nb_ctrlbots != "") && ([lsearch -exact $nb_ctrlbots [string tolower $frombot]] == -1))} {
        putlog "netbots: rejected command from $frombot (not a control bot)." ; return 0
      }
    }
  }
  set argstring [lindex $arg 2]
  switch -exact -- $command {
    "netbots" {nb_netbots $frombot $argstring}
    "netinfo" {nb_netinfo $frombot $argstring}
    "rnetinfo" {nb_rnetinfo $frombot $argstring}
    "netshell" {nb_netshell $frombot $argstring}
    "rnetshell" {nb_rnetshell $frombot $argstring}
    "netserv" {nb_netserv $frombot $argstring}
    "rnetserv" {nb_rnetserv $frombot $argstring}
    "netpass" {nb_netpass $frombot $argstring}
    "rnetpass" {nb_rnetpass $frombot $argstring}
    "netsave" {nb_netsave $frombot $argstring}
    "nethash" {nb_nethash $frombot $argstring}
    "netsay" {nb_netsay $frombot $argstring}
    "netact" {nb_netact $frombot $argstring}
    "netnotc" {nb_netnotc $frombot $argstring}
    "netdump" {nb_netdump $frombot $argstring}
    "netnick" {nb_netnick $frombot $argstring}
    "netjump" {nb_netjump $frombot $argstring}
    "netjoin" {nb_netjoin $frombot $argstring}
    "netpart" {nb_netpart $frombot $argstring}
    "netchanset" {nb_netchanset $frombot $argstring}
    "netupdate" {
      if {!$nb_nudisable} {nb_netupdate $frombot $argstring}
    }
    "rnetupdate" {
      if {!$nb_nudisable} {nb_rnetupdate $frombot $argstring}
    }
    "autopass" {nb_autopass $frombot $argstring}
    "rbroadcast" {nb_broadcast $frombot $argstring}
  }
  if {[info exists nb_netcmds($command)]} {
    $nb_netcmds($command) $frombot $argstring
  }
  return 0
}

proc nb_checkowner {hand idx} {
  global nb_owner owner
  if {(($nb_owner) && ([lsearch -exact [split [string tolower $owner] ", "] [string tolower $hand]] == -1))} {
    putidx $idx "What?  You need '.help'" ; return 0
  }
  return 1
}

proc nb_chattr {hand idx arg} {
  global nb_flag nb_owner owner
  if {[string match *$nb_flag* [lindex [split $arg] 1]]} {
    if {((![matchattr $hand n]) || (($nb_owner) && ([lsearch -exact [split [string tolower $owner] ", "] [string tolower $hand]] == -1)))} {
      putidx $idx "You do not have access to add/remove the netbot (+$nb_flag) flag."
      return
    }
  }
  *dcc:chattr $hand $idx $arg
}

proc nb_randpass {} {
  set l [expr 8 + [rand 4]]
  for {set x 0} {$x < $l} {incr x} {
    append randpass [string index "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890" [rand 62]]
  }
  return $randpass
}

proc nb_randomise {initial list} {
  set randlist ""
  if {$initial != "*"} {
    foreach element [split $initial] {
      lappend randlist $element
    }
  }
  while {[llength $list] > 0} {
    set random [rand [llength $list]]
    lappend randlist [lindex $list $random]
    set list [lreplace $list $random $random]
  }
  return $randlist
}

proc putwrap {idx ind text} {
# This is probably the ugliest proc I've even written :-)
  set string $text
  set length [expr 69 - $ind]
  set ind [string range "                                        " 0 [expr $ind - 1]]
  while {$string != ""} {
    if {(([string length $string] > [expr $length + 1]) && ([set last [string last " " [string range $string 0 $length]]] != -1))} {
      set first [expr $last + 1]
      set last [expr $last - 1]
    } else {
      set last $length
      set first [expr $length + 1]
    }
    putidx $idx $ind[string range $string 0 $last]
    set string [string trimleft [string range $string $first end]]
  }
  return
}

proc nb_plural {word llength} {
  if {$llength != 1} {return [join [split "$word s"] ""]}
  return $word
}

proc nb_flag2word {flags} {
  set words ""
  foreach flag [split $flags ""] {
    switch -exact -- $flag {
      "a" {lappend words "auto-ops"}
      "b" {lappend words "bots"}
      "c" {lappend words "commons"}
      "d" {lappend words "deops"}
      "f" {lappend words "friends"}
      "g" {lappend words "auto-voices"}
      "h" {lappend words "high-lights"}
      "j" {lappend words "janitors"}
      "k" {lappend words "auto-kicks"}
      "m" {lappend words "masters"}
      "n" {lappend words "owners"}
      "o" {lappend words "ops"}
      "p" {lappend words "party-lines"}
      "t" {lappend words "botnet masters"}
      "u" {lappend words "unshared"}
      "v" {lappend words "voices"}
      "w" {lappend words "wasop-tests"}
      "x" {lappend words "xfers"}
      "default" {lappend words "+$flag users"}
    }
  }
  if {[llength $words] > 1} {
    return "[join [lrange $words 0 [expr [llength $words] - 2]] ", "] and [lindex $words [expr [llength $words] - 1]]"
  }
  return $words
}

proc nb_killtimer {cmd} {
  set n 0
  regsub -all -- {\[} $cmd {\[} cmd ; regsub -all -- {\]} $cmd {\]} cmd
  foreach tmr [timers] {
    if {[string match $cmd [join [lindex $tmr 1]]]} {
      killtimer [lindex $tmr 2]
      incr n
    }
  }
  return $n
}

proc nb_killutimer {cmd} {
  set n 0
  regsub -all -- {\[} $cmd {\[} cmd ; regsub -all -- {\]} $cmd {\]} cmd
  foreach tmr [utimers] {
    if {[string match $cmd [join [lindex $tmr 1]]]} {
      killutimer [lindex $tmr 2]
      incr n
    }
  }
  return $n
}

proc nb_randport {servers} {
# This proc is based on that used in randomservers.tcl by FireEgl
  set tmpservers ""
  foreach server $servers {
    set ports [lindex [split $server :] 1]
    set pass [lindex [split $server :] 2]
    if {$ports != ""} {
      set portlist ""
      foreach group [split $ports ,] {
        set lowport [lindex [split $group -] 0]
        set highport [lindex [split $group -] 1]
        if {$highport == ""} {
          lappend portlist $lowport
        } else {
          while {$highport >= $lowport} {
            lappend portlist $lowport
            incr lowport
          }
        }
      }
      if {$pass != ""} {
        lappend tmpservers [lindex [split $server :] 0]:[lindex $portlist [rand [llength $portlist]]]:$pass
      } else {
        lappend tmpservers [lindex [split $server :] 0]:[lindex $portlist [rand [llength $portlist]]]
      }
    } else {
      lappend tmpservers $server
    }
  }
  return $tmpservers
}

proc nb_duration {time} {
# This proc is based on that used in Bass's Seen script (http://bseen.tclslave.net)
  set years 0 ; set days 0 ; set hours 0 ; set mins 0
  if {$time < 60} {return "< 1 min"}
  if {$time >= 31536000} {
    set years [expr int([expr $time/31536000])] ; set time [expr $time - [expr 31536000*$years]]
  }
  if {$time >= 86400} {
    set days [expr int([expr $time/86400])] ; set time [expr $time - [expr 86400*$days]]
  }
  if {$time >= 3600} {
    set hours [expr int([expr $time/3600])] ; set time [expr $time - [expr 3600*$hours]]
  }
  if {$time >= 60} {
    set mins [expr int([expr $time/60])]
  }
  if {$years == 0} {
    set output ""
  } elseif {$years == 1} {
    set output "1 yr,"
  } else {
    set output "$years yrs,"
  }
  if {$days == 1} {
    lappend output "1 day,"
  } elseif {$days > 1} {
    lappend output "$days days,"
  }
  if {$hours == 1} {
    lappend output "1 hr,"
  } elseif {$hours > 1} {
    lappend output "$hours hrs,"
  }
  if {$mins == 1} {
    lappend output "1 min"
  } elseif {$mins > 1} {
    lappend output "$mins mins"
  }
  return [string trimright [join $output] ", "]
}

if {[info exists nb_ctrlbot]} {
  unset nb_ctrlbot
}

if {[array exists nb_ctrlbots]} {
  proc nb_setctrlbots {} {
    global botnet-nick nb_ctrlbot nb_ctrlbots
    foreach botgroup [array names nb_ctrlbots] {
      if {$botgroup == "*"} {
        set nb_ctrlbot(*) [split $nb_ctrlbots(*)]
      } else {
        set nb_ctrlbots($botgroup) [string tolower $nb_ctrlbots($botgroup)]
        set bots [split [string tolower $botgroup] ,]
        foreach bot $bots {
          if {$bot == [string tolower ${botnet-nick}]} {continue}
          set nb_ctrlbot($bot) [split $nb_ctrlbots($botgroup)]
        }
      }
    }
    if {![info exists nb_ctrlbot(*)]} {
      set nb_ctrlbot(*) ""
    }
    return
  }
  nb_setctrlbots
  rename nb_setctrlbots ""
} else {
  set nb_ctrlbots [split [string tolower $nb_ctrlbots]]
}

if {![info exists nb_control]} {
  set nb_control $nb_defctrl
}

if {(([info exists nb_servers]) && ([llength $nb_servers] >= 1))} {
  set servers [nb_randport [nb_randomise * $nb_servers]]
}

if {![info exists nb_getting]} {
  set nb_getting 0
}

if {![info exists nb_tmin]} {
  set nb_tmin [expr [rand 9] + 1]
}
if {![info exists nb_thour]} {
  set nb_thour [rand 24]
  if {[string length $nb_thour] == 1} {
    set nb_thour 0$nb_thour
  }
}
bind time - "2$nb_tmin $nb_thour * * *" nb_update
bind time - "1$nb_tmin $nb_thour * * *" nb_tautopass
if {((!$nb_autopass) || ($nb_flag == "b") || ($nb_flag == "all"))} {
  unbind time - "1$nb_tmin $nb_thour * * *" nb_tautopass
  rename nb_tautopass ""
}
bind filt - * nb_cmdcast
if {!$nb_cmdcast} {
  unbind filt - * nb_cmdcast
  rename nb_cmdcast ""
}
if {(($nb_flag != "b") && ($nb_flag != "all"))} {
  unbind dcc m|m chattr *dcc:chattr
  bind dcc m|m chattr nb_chattr
} else {
  catch {unbind dcc m|m chattr nb_chattr}
  rename nb_chattr ""
}
bind bot - nb nb_gotcmd
bind dcc - nethelp nb_dccnethelp
bind dcc n components nb_dcccomponents
bind dcc n netbots nb_dccnetbots
bind dcc n netcontrol nb_dccnetcontrol
bind dcc n control nb_dccnetcontrol
bind dcc n netinfo nb_dccnetinfo
bind dcc n netshell nb_dccnetshell
bind dcc n netserv nb_dccnetserv
bind dcc n netpass nb_dccnetpass
bind dcc n netsave nb_dccnetsave
bind dcc n nethash nb_dccnethash
bind dcc n netsay nb_dccnetsay
bind dcc n netact nb_dccnetact
bind dcc n netnotice nb_dccnetnotc
bind dcc n netdump nb_dccnetdump
bind dcc n netnick nb_dccnetnick
bind dcc n netjump nb_dccnetjump
bind dcc n netjoin nb_dccnetjoin
bind dcc n netpart nb_dccnetpart
bind dcc n netchanset nb_dccnetchanset
bind dcc n netupdate nb_dccnetupdate
if {$nb_nudisable} {unbind dcc n netupdate nb_dccnetupdate}
bind chon - * nb_chon
bind chof - * nb_chof
if {!$nb_chon} {unbind chof - * nb_chof}
if {$numversion >= 1032800} {
  bind evnt - sighup nb_sighup ; bind evnt - sigterm nb_sigterm
}
catch {bind raw - 001 nb_svconnect}
catch {bind raw - 465 nb_svkline}

if {![info exists nb_components]} {
  set nb_components "netbots"
}
set nb_scripts $nb_script

if {[info exists nb_component]} {
  foreach nb_lcomponent [lsort [array names nb_component]] {
    if {!$nb_component($nb_lcomponent)} {continue}
    set nb_lsource $nb_dir/$nb_lcomponent.tcl
    lappend nb_scripts $nb_lsource
    if {(([file readable $nb_lsource]) && ([file isfile $nb_lsource]))} {
      if {[catch {source $nb_lsource} nb_error]} {
        putlog "*** WARNING: error found when loading $nb_lsource:"
        putlog $nb_error
        utimer 3 [split "nb_berror $nb_lsource"]
        return
      }
      if {[lsearch -exact $nb_components $nb_lcomponent] == -1} {
        lappend nb_components $nb_lcomponent
      }
    } else {
      putlog "Unable to load component '$nb_lcomponent' - $nb_lsource is not readable or doesn't exist."
    }
  }
}
if {[info exists nb_lcomponent]} {
  unset nb_lcomponent
}
if {[info exists nb_lsource]} {
  unset nb_lsource
}

putlog "Loaded components: [join [set nb_components [lsort $nb_components]] ", "]"

return
